@@ -54,13 +54,14 @@ inheritance_alias = {}
 
 # theme
 html_theme = "pydata_sphinx_theme"
-# html_logo = "_static/ipyforcegraph.svg"
-# html_favicon = "_static/favicon.ico"
+html_logo = "_static/wordmark.svg"
+html_favicon = "_static/logo.svg"
 
 html_theme_options = {
     "github_url": PROJ["urls"]["Source"],
     "use_edit_page_button": True,
     "show_toc_level": 1,
+    "icon_links": [],
 }
 html_context = {
     "github_user": "jupyrdf",
