@@ -12,11 +12,7 @@
 
 import os
 import subprocess
-import tempfile
-import textwrap
-import typing
 from hashlib import sha256
-from pathlib import Path
 
 from doit import create_after
 from doit.action import CmdAction
@@ -33,6 +29,7 @@ from scripts import utils as U
 
 os.environ.update(
     BLACK_CACHE_DIR=str(P.BUILD / ".black"),
+    JUPYTER_PLATFORM_DIRS="1",
     MAMBA_NO_BANNER="1",
     NODE_OPTS="--max-old-space-size=4096",
     PIP_DISABLE_PIP_VERSION_CHECK="1",
@@ -43,15 +40,9 @@ os.environ.update(
 )
 
 DOIT_CONFIG = dict(
-    backend="sqlite3",
-    default_tasks=["binder"],
-    par_type="thread",
     reporter=reporter.GithubActionsReporter,
-    verbosity=2,
 )
 
-Paths = typing.List[Path]
-
 
 def _echo_ok(msg):
     def _echo():
@@ -83,75 +74,6 @@ def task_all():
     )
 
 
-def task_lock():
-    """create lockfiles from the binder environment and CI excursions."""
-
-    if not P.USE_LOCK_ENV:
-        return
-
-    def _lock_comment(env_yamls: Paths) -> str:
-        comment = ""
-        for env_file in reversed(env_yamls):
-            comment += env_file.read_text(encoding="utf-8").strip() + "\n"
-        return textwrap.indent(comment, "# ")
-
-    def _needs_lock(lockfile: Path, env_yamls: Paths) -> bool:
-        if not lockfile.exists():
-            return True
-        lock_text = lockfile.read_text(encoding="utf-8")
-        comment = _lock_comment(env_yamls)
-        return comment not in lock_text
-
-    def _lock_one(lockfile: Path, env_yamls: typing.List[Path]) -> None:
-        if not _needs_lock(lockfile, env_yamls):
-            print(f"lockfile up-to-date: {lockfile}")
-            return
-
-        lock_args = ["conda-lock", "--kind=explicit"]
-        comment = _lock_comment(env_yamls)
-        for env_file in reversed(env_yamls):
-            lock_args += ["--file", env_file]
-        platform = next(p.stem for p in env_yamls if "subdir" in p.parent.name)
-        lock_args += ["--platform", platform]
-
-        if P.LOCK_HISTORY.exists():
-            lock_args = [*P.IN_LOCK_ENV, *lock_args]
-        elif not P.HAS_CONDA_LOCK:
-            print(
-                "Can't bootstrap lockfiles without `conda-lock`, please:\n\n\t"
-                "mamba install -c conda-forge conda-lock\n\n"
-                "and re-run `doit lock`"
-            )
-            return False
-
-        with tempfile.TemporaryDirectory() as td:
-            tdp = Path(td)
-            tmp_lock = tdp / f"conda-{platform}.lock"
-            subprocess.check_call(list(map(str, lock_args)), cwd=td)
-            raw = tmp_lock.read_text(encoding="utf-8").split(P.EXPLICIT)[1].strip()
-
-        lockfile.parent.mkdir(exist_ok=True, parents=True)
-        lockfile.write_text("\n".join([comment, P.EXPLICIT, raw, ""]), encoding="utf-8")
-
-    for env_yamls in P.ENV_MATRIX:
-        file_dep = [*env_yamls]
-        lock_name = "_".join([f"{p.stem}" for p in env_yamls])
-        lockfile = P.LOCKS / f"{lock_name}.conda.lock"
-
-        if not _needs_lock(lockfile, env_yamls):
-            continue
-
-        if P.LOCK_ENV_YAML not in env_yamls:
-            file_dep += [P.LOCK_HISTORY]
-
-        yield dict(
-            name=lock_name,
-            actions=[(_lock_one, [lockfile, env_yamls])],
-            file_dep=file_dep,
-            targets=[lockfile],
-        )
-
-
 def _ok(task, ok):
     task.setdefault("targets", []).append(ok)
     task["actions"] = [
@@ -159,7 +81,7 @@ def _ok(task, ok):
         *task["actions"],
         lambda: [
             ok.parent.mkdir(exist_ok=True, parents=True),
-            ok.write_text("ok", encoding="utf-8"),
+            ok.write_text("ok", **P.UTF8),
             True,
         ][-1],
     ]
@@ -240,22 +162,41 @@ def task_binder():
 
 
 def task_env():
-    """prepare project envs"""
-    yield dict(
-        name="sync:docs:binder",
-        file_dep=[P.DOCS_ENV_YAML],
-        targets=[P.BINDER_ENV_YAML],
-        actions=[
-            (
-                U.replace_between_patterns,
-                [P.DOCS_ENV_YAML, P.BINDER_ENV_YAML, "### docs/environment.yml ###"],
-            )
-        ],
-    )
-
-    if not P.USE_LOCK_ENV:
+    """ensure environment reproducibility."""
+    if P.CI or P.IN_RTD or P.IN_BINDER:
         return
 
+    for spec_path in P.ENV_SPECS.glob("*.yml"):
+        if spec_path.name.startswith("_"):
+            continue
+        spec = U.safe_load(spec_path)
+        for platform in P.ALL_PLATFORMS:
+            for stack in U.get_spec_stacks(spec_path, platform):
+                yml_target = spec.get("_target")
+                task = dict(file_dep=stack)
+                if yml_target:
+                    target = spec_path.parent / yml_target
+                    yield dict(
+                        doc=f"build environment.yml for {target.parent.name}",
+                        actions=[(U.merge_envs, [target, stack])],
+                        name=f"yml:{spec_path.stem}",
+                        targets=[target.resolve()],
+                        **task,
+                    )
+                else:
+                    lockfile = "_".join(
+                        [platform, stack[0].stem]
+                        + [s.stem for s in stack if s.parent != P.ENV_SPECS]
+                    )
+                    target = P.LOCKS / f"{lockfile}.conda.lock"
+                    yield dict(
+                        doc=f"build lockfile for {lockfile.replace('_', ' ')}",
+                        name=f"lock:{lockfile}",
+                        actions=[(U.lock_one, [platform, target, stack])],
+                        targets=[target],
+                        **task,
+                    )
+
     yield dict(
         name="lock",
         file_dep=[P.LOCK_LOCKFILE],
@@ -272,7 +213,6 @@ def task_env():
         targets=[P.HISTORY],
         actions=[
             [*P.MAMBA_CREATE, P.ENV, "--file", P.LOCKFILE],
-            [*P.IN_ENV, *P.PIP, "install", "--no-deps", *P.LITE_SPEC],
         ],
     )
 
@@ -322,10 +262,13 @@ def task_setup():
         # ancient sphinx_rtd_theme wants ancient docutils
         py_actions += [[*P.IN_ENV, *P.PIP, "check"]]
 
+    if P.CI:
+        print("setup:py actions")
+        print(py_actions)
+
     py_task = _ok(
         dict(
             name="py",
-            uptodate=[config_changed({"artifact": P.INSTALL_ARTIFACT})],
             file_dep=file_dep,
             actions=py_actions,
         ),
@@ -445,7 +388,7 @@ def task_build():
 
         output = "\n".join(lines)
         print(output)
-        P.SHA256SUMS.write_text(output)
+        P.SHA256SUMS.write_text(output, **P.UTF8)
 
     yield dict(
         name="hash",
@@ -487,6 +430,8 @@ def task_pytest():
 
 def task_test():
     """run all the notebooks"""
+    if P.IN_BINDER:
+        return
 
     def _nb_test(nb):
         def _test():
@@ -506,6 +451,8 @@ def task_test():
             ]
             return CmdAction(args, env=env, shell=False)
 
+        html = P.BUILD_NBHTML / nb.name.replace(".ipynb", ".html")
+
         file_dep = [
             *P.ALL_PY_SRC,
             *P.EXAMPLE_IPYNB,
@@ -519,26 +466,37 @@ def task_test():
         return dict(
             name=f"nb:{nb.name}".replace(" ", "_").replace(".ipynb", ""),
             file_dep=file_dep,
-            actions=[_test()],
-            targets=[P.BUILD_NBHTML / nb.name.replace(".ipynb", ".html")],
+            actions=[
+                (U.clean_some, [html]),
+                _test(),
+                (U.html_expect_xpath_matches, [html]),
+            ],
+            targets=[html],
         )
 
     for nb in P.EXAMPLE_IPYNB:
         yield _nb_test(nb)
 
     for robot_template in P.ATEST.rglob("*.j2"):
-        robot_out = robot_template.parent / robot_template.name.replace(".j2", "")
-        yield dict(
-            name=f"atest:template:{robot_template.relative_to(P.ATEST)}",
-            actions=[(U.template_one, [robot_template, robot_out])],
-            file_dep=[robot_template],
-            targets=[robot_out],
-        )
+        for graph_class in P.PY_GRAPH_CLASSES:
+            name = robot_template.name.replace(".j2", "").replace("GRAPH", graph_class)
+            robot_out = robot_template.parent / name
+            yield dict(
+                name=f"atest:template:{robot_out.relative_to(P.ATEST)}",
+                actions=[
+                    (
+                        U.template_one,
+                        [robot_template, robot_out, {"graph_class": graph_class}],
+                    )
+                ],
+                file_dep=[robot_template],
+                targets=[robot_out],
+            )
 
     def _pabot_logs():
         for robot_out in sorted(P.ATEST_OUT.rglob("robot_*.out")):
             print(f"\n[{robot_out.relative_to(P.ROOT)}]")
-            print(robot_out.read_text() or "<EMPTY>")
+            print(robot_out.read_text(**P.UTF8) or "<EMPTY>")
 
     yield dict(
         name="atest",
@@ -560,7 +518,7 @@ def task_test():
 
 def task_lint():
     """format all source files"""
-    if P.TESTING_IN_CI:
+    if P.TESTING_IN_CI or P.IN_BINDER:
         return
 
     yield _ok(
@@ -650,6 +608,15 @@ def task_lint():
             P.OK_NBLINT[nb.name],
         )
 
+    yield _ok(
+        dict(
+            name="dos2unix",
+            file_dep=[*P.ALL_DOS2UNIX, *[*P.OK_NBLINT.values()]],
+            actions=[U.fix_windows_line_endings],
+        ),
+        P.OK_DOS2UNIX,
+    )
+
     yield _ok(
         dict(
             name="robot",
@@ -669,7 +636,7 @@ def task_lint():
         P.OK_ROBOT_LINT,
     )
 
-    index_src = P.EXAMPLE_INDEX.read_text(encoding="utf-8")
+    index_src = P.EXAMPLE_INDEX.read_text(**P.UTF8)
 
     def _make_index_check(ex):
         def _check():
@@ -857,7 +824,7 @@ def _make_spellcheck(dep, html):
             print("Unrecognized words in", dep)
             print(fail_str)
             fail_path.parent.mkdir(exist_ok=True, parents=True)
-            fail_path.write_text(fail_str, encoding="utf-8")
+            fail_path.write_text(fail_str, **P.UTF8)
 
     return _ok(
         dict(
@@ -875,7 +842,7 @@ def _all_spell():
     for path in sorted(P.ALL_SPELL.parent.rglob("*.fail")):
         if path == P.ALL_SPELL:
             continue
-        all_fail += path.read_text(encoding="utf-8").strip().splitlines()
+        all_fail += path.read_text(**P.UTF8).strip().splitlines()
 
     all_fail_str = "\n".join(sorted(set(all_fail)))
 
@@ -883,7 +850,7 @@ def _all_spell():
         print("ALL Unrecognized words")
         print(all_fail_str)
 
-    P.ALL_SPELL.write_text(all_fail_str, encoding="utf-8")
+    P.ALL_SPELL.write_text(all_fail_str, **P.UTF8)
 
     return len(all_fail) == 0
 
@@ -917,7 +884,8 @@ def task_checkdocs():
                         "--check-links-cache",
                         *["--check-links-cache-name", P.DOCS_LINKS],
                         # TODO: relax these once published
-                        *["-k", "not (edit or rtfd or pypi)"],
+                        "--check-links-ignore",
+                        "https://",
                         "--links-ext=html",
                         *file_dep,
                     ],
@@ -931,8 +899,19 @@ def task_checkdocs():
 
     spell_tasks = []
 
+    yield _ok(
+        dict(
+            name="dictionary",
+            doc="ensure dictionary is unique and sorted",
+            file_dep=[P.DICTIONARY],
+            actions=[(U.sort_unique, [P.DICTIONARY])],
+        ),
+        P.OK_DICTIONARY,
+    )
+
     for dep in file_dep:
         task = _make_spellcheck(dep, html)
+        task["file_dep"] += [P.OK_DICTIONARY]
         spell_tasks += [f"""checkdocs:{task["name"]}"""]
         yield task
 
