@@ -2,9 +2,9 @@
  * Copyright (c) 2023 ipyforcegraph contributors.
  * Distributed under the terms of the Modified BSD License.
  */
-import d3Force3d from 'd3-force-3d';
+import type d3Force3d from 'd3-force-3d';
 import type { GraphData, LinkObject, NodeObject } from 'force-graph';
-import type { WebGLRenderer } from 'three';
+import type THREE from 'three';
 
 import type { ISignal } from '@lumino/signaling';
 
@@ -12,6 +12,8 @@ import type { DOMWidgetView } from '@jupyter-widgets/base';
 
 import PKG from '../package.json';
 
+import type { ForceGraphModel } from './widgets';
+
 export const NAME = PKG.name;
 export const VERSION = PKG.version;
 export const EMOJI = '🕸️';
@@ -58,26 +60,30 @@ export enum EUpdate {
   Reheat = 1 << 0,
   Cosmetic = 1 << 1,
   Render = 1 << 2,
+  Behavior = 1 << 3,
 }
 export type TUpdateKind = void | number;
 
 export interface IBehave {
   updateRequested: ISignal<IBehave, TUpdateKind>;
+  extraColumns?: IExtraColumns;
   // link
   getLinkColor?(options: ILinkBehaveOptions): string | null;
-  getLinkWidth?(options: ILinkBehaveOptions): string | null;
+  getLinkWidth?(options: ILinkBehaveOptions): number | null;
   getLinkLabel?(options: ILinkBehaveOptions): string | null;
   getLinkDirectionalArrowColor?(options: ILinkBehaveOptions): string | null;
-  getLinkDirectionalArrowLength?(options: ILinkBehaveOptions): string | null;
-  getLinkDirectionalArrowRelPos?(options: ILinkBehaveOptions): string | null;
+  getLinkDirectionalArrowLength?(options: ILinkBehaveOptions): number | null;
+  getLinkDirectionalArrowRelPos?(options: ILinkBehaveOptions): number | null;
   getLinkDirectionalParticleColor?(options: ILinkBehaveOptions): string | null;
-  getLinkDirectionalParticleSpeed?(options: ILinkBehaveOptions): string | null;
-  getLinkDirectionalParticleWidth?(options: ILinkBehaveOptions): string | null;
-  getLinkDirectionalParticles?(options: ILinkBehaveOptions): string | null;
+  getLinkDirectionalParticleSpeed?(options: ILinkBehaveOptions): number | null;
+  getLinkDirectionalParticleWidth?(options: ILinkBehaveOptions): number | null;
+  getLinkDirectionalParticles?(options: ILinkBehaveOptions): number | null;
   // node
   getNodeColor?(options: INodeBehaveOptions): string | null;
   getNodeLabel?(options: INodeBehaveOptions): string | null;
-  getNodeSize?(options: INodeBehaveOptions): string | null;
+  getNodeSize?(options: INodeBehaveOptions): number | null;
+  getNodeCanvasObject?(options: INodeCanvasBehaveOptions): any;
+  getNodeThreeObject?(options: INodeThreeBehaveOptions): THREE.Object3D | null;
   // evented
   onNodeClick?(options: INodeEventBehaveOptions): boolean;
   onLinkClick?(options: ILinkEventBehaveOptions): boolean;
@@ -103,6 +109,8 @@ export const ALL_NODE_METHODS = [
   'getNodeLabel',
   'getNodeColor',
   'getNodeSize',
+  'getNodeCanvasObject',
+  'getNodeThreeObject',
   'onNodeClick',
 ];
 export type TNodeBehaveMethod = (typeof ALL_NODE_METHODS)[number];
@@ -123,8 +131,18 @@ export interface INodeBehaveOptions extends IBehaveOptions {
   node: NodeObject;
 }
 
+export interface INodeCanvasBehaveOptions extends INodeBehaveOptions {
+  context: CanvasRenderingContext2D;
+  globalScale: number;
+}
+
+export interface INodeThreeBehaveOptions extends INodeBehaveOptions {
+  iframeClasses: Record<string, any>;
+}
+
 export interface INodeEventBehaveOptions extends INodeBehaveOptions {
   event: MouseEvent;
+  index: number;
 }
 
 export interface ILinkBehaveOptions extends IBehaveOptions {
@@ -138,7 +156,7 @@ export interface ILinkEventBehaveOptions extends ILinkBehaveOptions {
 
 export interface IRenderOptions extends IBehaveOptions {
   context2d?: CanvasRenderingContext2D;
-  renderer3d?: WebGLRenderer;
+  renderer3d?: THREE.WebGLRenderer;
   globalScale?: number;
   time?: number;
 }
@@ -148,6 +166,7 @@ export interface IHasGraph<T = any> extends DOMWidgetView {
   source: ISource;
   rendered: Promise<void>;
   wrapFunction: (fn: Function) => Function;
+  model: ForceGraphModel;
 }
 
 export interface ISource {
@@ -173,3 +192,39 @@ export interface IForce {
 export type TSelectedSet = Set<string | number>;
 
 export const emptyArray = Object.freeze([]);
+
+export interface IDynamicCallable {
+  (...args: any): string;
+}
+
+export type TCustomAction = 'reheat';
+
+export interface IActionMessage {
+  action: TCustomAction;
+}
+
+/**
+ * Strings that should be interpreted as `false` after being lowercased and trimmed.
+ */
+export const FALSEY = Object.freeze([
+  '',
+  '()',
+  '[]',
+  '{}',
+  '0.0',
+  '0',
+  'false',
+  'nan',
+  'none',
+  'null',
+]);
+
+export enum ECoerce {
+  boolish = 'boolean',
+  numeric = 'number',
+}
+
+export interface IExtraColumns {
+  nodes: string[];
+  links: string[];
+}
