@@ -33,11 +33,13 @@ import {
   EMOJI,
   EMPTY_GRAPH_DATA,
   EUpdate,
+  IActionMessage,
   IBehave,
   IHasGraph,
   ILinkBehaveOptions,
   ILinkEventBehaveOptions,
   INodeBehaveOptions,
+  INodeCanvasBehaveOptions,
   INodeEventBehaveOptions,
   IRenderOptions,
   ISource,
@@ -52,7 +54,7 @@ import {
   WIDGET_DEFAULTS,
   emptyArray,
 } from '../../tokens';
-import { ForceBehaviorModel, GraphForcesModel } from '../behaviors';
+import { DAGBehaviorModel, FacetedForceModel, GraphForcesModel } from '../behaviors';
 
 export class ForceGraphModel extends DOMWidgetModel {
   static model_name = 'ForceGraphModel';
@@ -87,10 +89,10 @@ export class ForceGraphModel extends DOMWidgetModel {
   initialize(attributes: ObjectHash, options: IBackboneModelOptions): void {
     super.initialize(attributes, options);
     this.on('change:behaviors', this.onBehaviorsChange, this);
-    this.onBehaviorsChange();
+    void this.onBehaviorsChange();
   }
 
-  onBehaviorsChange(): void {
+  async onBehaviorsChange(): Promise<void> {
     if (!this._behaviorsChanged) {
       this._behaviorsChanged = new Signal(this);
       this._linkBehaviorsByMethod = new Map();
@@ -135,6 +137,7 @@ export class ForceGraphModel extends DOMWidgetModel {
         this._forceBehaviors.push(behavior);
       }
     }
+
     this._behaviorsChanged.emit(void 0);
   }
 
@@ -202,6 +205,7 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
 
   protected _rendered: PromiseDelegate<void>;
   protected _iframe: HTMLIFrameElement | null;
+  protected _iframeClasses: Record<string, any>;
 
   get source(): ISource {
     return this.model.get('source');
@@ -229,9 +233,27 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
 
     this.model.behaviorsChanged.connect(this.onBehaviorsChange, this);
     this.luminoWidget.disposed.connect(this.onDisposed, this);
-
+    this.model.on('msg:custom', this.handleMessage, this);
     this.onSourceChange();
-    this.onBehaviorsChange();
+    void this.onBehaviorsChange();
+  }
+
+  handleMessage(message: IActionMessage): void {
+    const graph = this.graph as ForceGraphInstance;
+
+    if (!graph) {
+      console.warn(`${EMOJI} graph was not yet initialized, discarding`, message);
+      return;
+    }
+
+    switch (message.action) {
+      case 'reheat':
+        graph.d3ReheatSimulation();
+        break;
+      default:
+        const exhaustiveCheck: never = message.action;
+        console.error(`${EMOJI} Unhandled custom action: ${exhaustiveCheck}`);
+    }
   }
 
   onDisposed() {
@@ -265,7 +287,10 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
     const iframe = event.currentTarget as HTMLIFrameElement;
     const { contentWindow } = iframe;
 
-    const graph: ForceGraphInstance = (contentWindow as any).init();
+    const initResult = (contentWindow as any).init();
+
+    const graph: ForceGraphInstance = initResult.graph;
+    this._iframeClasses = initResult.iframeClasses;
     this.graph = graph as any;
     contentWindow.addEventListener('resize', this.onWindowResize);
     this._rendered.resolve(void 0);
@@ -279,24 +304,36 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
     graph.height(contentWindow.innerHeight);
   };
 
-  protected async getJsUrl(): Promise<string> {
-    return (
-      await import(
-        '!!file-loader!../../../node_modules/force-graph/dist/force-graph.js'
-      )
-    ).default as any;
+  protected async getJsUrls(): Promise<string[]> {
+    return [
+      (
+        await import(
+          '!!file-loader!../../../node_modules/force-graph/dist/force-graph.js'
+        )
+      ).default as any,
+    ];
   }
 
   protected get graphJsClass(): string {
     return 'ForceGraph';
   }
 
+  protected get extraJsClasses(): string {
+    return '{}';
+  }
+
   protected async getIframeSource(): Promise<string> {
-    let url = await this.getJsUrl();
+    let urls = await this.getJsUrls();
+
+    let scripts = '';
+
+    for (const url of urls) {
+      scripts += `<script src="${url}"></script>\n`;
+    }
 
     let src = `
       <head>
-        <script src="${url}"></script>
+        ${scripts}
         <style>
           body, #main {
             overflow: hidden;
@@ -320,10 +357,15 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
           window.init = (args) => {
             const div = document.createElement('div');
             document.body.appendChild(div);
-            return ${this.graphJsClass}(args || {})(div);
+            return {
+              graph: ${this.graphJsClass}(args || {})(div),
+              iframeClasses: ${this.extraJsClasses}
+            };
           }
           window.wrapFunction = (fn) => {
-            return (...args) => fn(...args);
+            return (...args) => {
+              return fn(...args);
+            };
           }
         </script>
       </body>
@@ -339,9 +381,9 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
     (this.graph as any).graphData(graphData);
   }
 
-  wrapFunction(fn: Function) {
+  wrapFunction = (fn: Function) => {
     return (this._iframe.contentWindow as any).wrapFunction(fn);
-  }
+  };
 
   onSourceChange(change?: any) {
     const { source, previousSource } = this;
@@ -356,7 +398,23 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
     }
   }
 
+  protected async ensureAllFacets() {
+    let facetPromises: Promise<void>[] = [];
+
+    for (const behavior of this.model.behaviors) {
+      // TOOD: remove the any
+      if ((behavior as any).ensureFacets) {
+        facetPromises.push((behavior as any).ensureFacets());
+      }
+    }
+
+    if (facetPromises.length) {
+      await Promise.all(facetPromises);
+    }
+  }
+
   protected async postUpdate(caller?: any, kind?: TUpdateKind): Promise<void> {
+    await this.displayed;
     await this.rendered;
     const graph = this.graph as ForceGraphInstance;
     if (!graph) {
@@ -364,6 +422,13 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
       return;
     }
 
+    if (kind && EUpdate.Behavior === (kind & EUpdate.Behavior)) {
+      await this.model.onBehaviorsChange();
+      return;
+    }
+
+    await this.ensureAllFacets();
+
     const {
       backgroundColor,
       defaultLinkColor,
@@ -471,6 +536,7 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
   protected getForceUpdate() {
     const graph = this.graph as ForceGraphInstance;
     let needsPost: TAnyForce[] = [];
+
     for (let simBehavior of this.model.forceBehaviors) {
       const { warmupTicks, cooldownTicks, alphaDecay, alphaMin, velocityDecay } =
         simBehavior;
@@ -482,12 +548,19 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
       graph.d3VelocityDecay(velocityDecay);
 
       for (let key in simBehavior.forces) {
-        let behavior: ForceBehaviorModel | null = simBehavior.forces[key];
+        let behavior: FacetedForceModel | null = simBehavior.forces[key];
         let force = behavior?.force || null;
         if (force && !behavior?.active) {
           force = null;
         }
-        graph.d3Force(key, force);
+
+        // DAG behavior is treated different compared to other pure D3 Forces.
+        if (behavior instanceof DAGBehaviorModel) {
+          (behavior as DAGBehaviorModel).refreshBehavior(graph);
+        } else {
+          graph.d3Force(key, force);
+        }
+
         if (force?.links) {
           needsPost.push(force);
         }
@@ -518,6 +591,13 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
 
   protected getOnRenderPostUpdate() {
     const graph = this.graph as ForceGraphInstance;
+
+    graph.nodeCanvasObject(
+      this.model.nodeBehaviorsForMethod('getNodeCanvasObject').length
+        ? this.wrapFunction(this.getNodeCanvasObject)
+        : null
+    );
+
     graph.onRenderFramePost(
       this.model.graphBehaviorsForMethod('onRender').length
         ? this.wrapFunction(this.onRender)
@@ -637,6 +717,30 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
     return this.getComposedNodeAttr(node, 'getNodeSize', this.model.defaultNodeSize);
   };
 
+  protected getNodeCanvasObject = (
+    node: NodeObject,
+    context: CanvasRenderingContext2D,
+    globalScale: number
+  ): void => {
+    let value: string | null;
+    const graphData = (this.graph as ForceGraphInstance).graphData();
+    const options: INodeCanvasBehaveOptions = {
+      view: this,
+      context,
+      graphData,
+      node,
+      globalScale,
+    };
+
+    for (const behavior of this.model.nodeBehaviorsForMethod('getNodeCanvasObject')) {
+      let method = behavior.getNodeCanvasObject;
+      value = method.call(behavior, options);
+      if (value != null) {
+        break;
+      }
+    }
+  };
+
   getComposedNodeAttr(
     node: NodeObject,
     methodName: TNodeBehaveMethod,
@@ -669,6 +773,7 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
       graphData,
       event,
       node,
+      index: node['index'] != null ? node['index'] : graphData.nodes.indexOf(node),
     };
     for (const behavior of this.model.nodeBehaviorsForMethod('onNodeClick')) {
       shouldContinue = behavior.onNodeClick(options);
@@ -686,7 +791,7 @@ export class ForceGraphView<T = ForceGraphGenericInstance<ForceGraphInstance>>
       graphData,
       event,
       link,
-      index: graphData.links.indexOf(link),
+      index: link['index'] != null ? link['index'] : graphData.links.indexOf(link),
     };
     for (const behavior of this.model.linkBehaviorsForMethod('onLinkClick')) {
       shouldContinue = behavior.onLinkClick(options);
